//
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//                                                                      %
// Copyright (C) 2010, Liangping Li,J. J. Gomez-Hernandez, and Haiyan   %
// Zhou.  Universidad Politecnica de Valencia,All rights reserved.      %
//                                                                      %
// This program is distributed in the hope that they will be useful,    %
// but WITHOUT ANY WARRANTY.  No author or distributor accepts          %
// responsibility to anyone for the consequences of using them or for   %
// whether they serve any particular purpose or work at all, unless he  %
// says so in writing.  Everyone is granted permission to copy, modify  %
// and redistribute this program, but only under the condition          %
// that this notice and the above copyright notice remain intact.       %
//                                                                      %
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//-----------------------------------------------------------------------


//
//   Steady state groundwater flow modleing with full tensor
//                    by finite-difference 
//          ****************************************
//
//
// The program is executed with no command line arguments.  The user
// will be prompted for the name of a parameter file.  The parameter
// file is described in the documentation (see the example flowxyz.par)
//
// All input and output files have a GeoEAS format.  
//

#include<iostream>
#include<fstream>
#pragma warning(disable:4786)
#include<vector>
#include<cmath>
#include<string>
using namespace std;

void Read_para(int &ncol,int &nrow,int &nlay,float &f_dx,float &f_dy,float &f_dz,int &max_iter,float &accl,float &closure,
			   string &kxx_file,string &kyy_file,string &kzz_file,string &ihead_file,string &ibound_file,
			   string &head_file,string &qx_file,string &qy_file,string &qz_file,int &n_well,string &well_file,
			   string &dx_file,string &dy_file,string &dz_file,int &flag_dx,int &flag_dy,int &flag_dz,int &debug);

void Read_data(int ncol,int nrow,int nlay,string &kxx_file,string &kyy_file,string &kzz_file,
			   string &ihead_file,string &ibound_file,vector< vector< vector<float> > > &dx_3d,
			   vector< vector< vector<float> > > &dy_3d,vector< vector< vector<float> > > &dz_3d,
			   vector< vector< vector<float> > > &head,
			   vector< vector< vector<int> > > &ibo,vector< vector< vector<float> > > &cc_xx,
			   vector< vector< vector<float> > > &cc_xy,vector< vector< vector<float> > > &cc_xz,
			   vector< vector< vector<float> > > &cr_yx,vector< vector< vector<float> > > &cr_yy,
			   vector< vector< vector<float> > > &cr_yz,vector< vector< vector<float> > > &cv_zx,
			   vector< vector< vector<float> > > &cv_zy,vector< vector< vector<float> > > &cv_zz,
			   string &well_file,int n_well,vector<float>&well_x,vector<float>&well_y,vector<float>&well_z,
			   vector<float>&well_q, string &dx_file,string &dy_file,string &dz_file,float f_dx,float f_dy,float f_dz,
			   int flag_dx,int flag_dy,int flag_dz,int debug);

void Set_boundary(int ncol,int nrow,int nlay,string &kxx_file,string &kyy_file,string &kzz_file,
				vector< vector< vector<int> > > &ibo,
			   vector< vector< vector<float> > > &cc_xx,vector< vector< vector<float> > > &cc_xy,
			   vector< vector< vector<float> > > &cc_xz,vector< vector< vector<float> > > &cr_yx,
			   vector< vector< vector<float> > > &cr_yy,vector< vector< vector<float> > > &cr_yz,
			   vector< vector< vector<float> > > &cv_zx,vector< vector< vector<float> > > &cv_zy,
			   vector< vector< vector<float> > > &cv_zz,int debug);

void write_debug1(int nlay,int nrow,int ncol,string &ibound_file,vector< vector< vector<int> > > &ibo);
void write_debug2(int nlay,int nrow,int ncol,string &ihead_file,vector< vector< vector<float> > > &head);
void write_debug3(int nlay,int nrow,int ncol,string &kxx_file,vector< vector< vector<float> > > &cc_xy);

void Compute_abcs(int nlay,int nrow,int ncol,vector< vector< vector<float> > >dx,vector< vector< vector<float> > >dy,
			   vector< vector< vector<float> > >dz,vector< vector< vector<float> > > &cc_xx,
			   vector< vector< vector<float> > > &cc_xy,vector< vector< vector<float> > > &cc_xz,
			   vector< vector< vector<float> > > &cr_yx,vector< vector< vector<float> > > &cr_yy,
			   vector< vector< vector<float> > > &cr_yz,vector< vector< vector<float> > > &cv_zx,
			   vector< vector< vector<float> > > &cv_zy,vector< vector< vector<float> > > &cv_zz,
			   vector< vector< vector<float> > > &A,vector< vector< vector<float> > > &B,
			   vector< vector< vector<float> > > &C,vector< vector< vector<float> > > &D,
			   vector< vector< vector<float> > > &E,vector< vector< vector<float> > > &F,
			   vector< vector< vector<float> > > &G,vector< vector< vector<float> > > &H,
			   vector< vector< vector<float> > > &I,vector< vector< vector<float> > > &J,
			   vector< vector< vector<float> > > &K,vector< vector< vector<float> > > &L,
			   vector< vector< vector<float> > > &M,vector< vector< vector<float> > > &N,
			   vector< vector< vector<float> > > &O,vector< vector< vector<float> > > &P,
			   vector< vector< vector<float> > > &Q,vector< vector< vector<float> > > &R,
			   vector< vector< vector<float> > > &S,vector< vector< vector<float> > > &RHS,
			   vector< vector< vector<int> > > ibo,int debug);

void Fill_outside_ring(int nlay,int nrow,int ncol,vector< vector< vector<float> > > &head);

bool One_iteration(int nlay,int nrow,int ncol,float max_head_change,float closure,float accl,
				  vector< vector< vector<float> > > &head,vector< vector< vector<int> > > ibo,
				  vector< vector< vector<float> > > &A,vector< vector< vector<float> > > &B,
				  vector< vector< vector<float> > > &C,vector< vector< vector<float> > > &D,
				  vector< vector< vector<float> > > &E,vector< vector< vector<float> > > &F,
				  vector< vector< vector<float> > > &G,vector< vector< vector<float> > > &H,
				  vector< vector< vector<float> > > &I,vector< vector< vector<float> > > &J,
				  vector< vector< vector<float> > > &K,vector< vector< vector<float> > > &L,
				  vector< vector< vector<float> > > &M,vector< vector< vector<float> > > &N,
				  vector< vector< vector<float> > > &O,vector< vector< vector<float> > > &P,
				  vector< vector< vector<float> > > &Q,vector< vector< vector<float> > > &R,
				  vector< vector< vector<float> > > &S,vector< vector< vector<float> > > &RHS,
				  int debug,bool CONVERGENCE,bool NO_CONVERGENCE);

bool Iterate(int max_iter,int nlay,int nrow,int ncol,int debug,float max_head_change,float closure,float accl,
			vector< vector< vector<float> > > &head,vector< vector< vector<float> > > &A,
			vector< vector< vector<float> > > &B,vector< vector< vector<float> > > &C,
			vector< vector< vector<float> > > &D,vector< vector< vector<float> > > &E,
			vector< vector< vector<float> > > &F, vector< vector< vector<float> > > &G,
			vector< vector< vector<float> > > &H,vector< vector< vector<float> > > &I,
			vector< vector< vector<float> > > &J, vector< vector< vector<float> > > &K,
			vector< vector< vector<float> > > &L, vector< vector< vector<float> > > &M,
			vector< vector< vector<float> > > &N,vector< vector< vector<float> > > &O,
			vector< vector< vector<float> > > &P,vector< vector< vector<float> > > &Q,
			vector< vector< vector<float> > > &R, vector< vector< vector<float> > > &S,
			vector< vector< vector<float> > > &RHS,
			vector< vector< vector<int> > > ibo,bool CONVERGENCE,bool NO_CONVERGENCE);

void Compute_flow(int nlay,int nrow,int ncol,vector< vector< vector<float> > >dx,vector< vector< vector<float> > >dy,
				  vector< vector< vector<float> > >dz,vector< vector< vector<float> > > &head,
				  vector< vector< vector<float> > > &cc_xx,vector< vector< vector<float> > > &cc_xy,
				  vector< vector< vector<float> > > &cc_xz,vector< vector< vector<float> > > &cr_yx,
				  vector< vector< vector<float> > > &cr_yy,vector< vector< vector<float> > > &cr_yz,
				  vector< vector< vector<float> > > &cv_zx,vector< vector< vector<float> > > &cv_zy,
				  vector< vector< vector<float> > > &cv_zz,vector< vector< vector<float> > > &qxx,
				  vector< vector< vector<float> > > &qyy,vector< vector< vector<float> > > &qzz,int debug);

void Print_flow(int nlay,int nrow,int ncol,string qx_file,float average_qx,vector< vector< vector<float> > > &qxx);

void Print_head(int nlay,int nrow,int ncol,string head_file,vector< vector< vector<float> > > &head,vector< vector< vector<int> > > ibo);

void Check_flow(int nlay,int nrow,int ncol,vector< vector< vector<float> > > &qxx,vector< vector< vector<float> > > &qyy,
		   vector< vector< vector<float> > > &qzz,vector< vector< vector<float> > > &RHS,
		   vector< vector< vector<float> > > &head,vector< vector< vector<float> > >dx,vector< vector< vector<float> > >dy,
			   vector< vector< vector<float> > >dz);

void modifyK_well(int nlay,int nrow,int ncol,string &kxx_file,string &kyy_file,string &kzz_file,
				  vector< vector< vector<float> > > &cc_xx,vector< vector< vector<float> > > &cc_xy,
				  vector< vector< vector<float> > > &cc_xz,vector< vector< vector<float> > > &cr_yx,
				  vector< vector< vector<float> > > &cr_yy,vector< vector< vector<float> > > &cr_yz,
				  vector< vector< vector<float> > > &cv_zx,vector< vector< vector<float> > > &cv_zy,
				  vector< vector< vector<float> > > &cv_zz,vector<float>&well_x,vector<float>&well_y,
				  vector<float>&well_z,int n_well,int debug);
void modify_RHS_well(int nlay,int nrow,int ncol,vector<float>&well_x,vector<float>&well_y,
				  vector<float>&well_z,vector<float>&well_q,int n_well,vector< vector< vector<float> > > &RHS);


void main()
{
	int ncol,nrow,nlay,max_iter,n_well,flag_dx,flag_dy,flag_dz,debug;
	float f_dx,f_dy,f_dz,accl,closure,max_head_change,average_qx,average_qy,average_qz ;
	max_head_change=0;average_qx=0;average_qy=0;average_qz=0;
	string kxx_file,kyy_file,kzz_file,ihead_file,ibound_file,head_file,qx_file,qy_file,qz_file,well_file,
		   dx_file,dy_file,dz_file;
	bool CONVERGENCE=true;
	bool NO_CONVERGENCE=false;

	Read_para(ncol,nrow,nlay,f_dx,f_dy,f_dz,max_iter,accl,closure,kxx_file,kyy_file,kzz_file,ihead_file,ibound_file,
			head_file,qx_file,qy_file,qz_file,n_well,well_file,dx_file,dy_file,dz_file,flag_dx,flag_dy,flag_dz,
			debug);

	vector< vector< vector<float> > > dx_3d(nlay+2, vector< vector<float> > (nrow+2, vector<float>(ncol+2,0)) );
	vector< vector< vector<float> > > dy_3d(nlay+2, vector< vector<float> > (nrow+2, vector<float>(ncol+2,0)) );
	vector< vector< vector<float> > > dz_3d(nlay+2, vector< vector<float> > (nrow+2, vector<float>(ncol+2,0)) );

	vector< vector< vector< int > > > ibo  (nlay+2, vector< vector< int > > (nrow+2, vector< int >(ncol+2,1)) );
	vector< vector< vector<float> > > cc_xx(nlay+1, vector< vector<float> > (nrow+1, vector<float>(ncol+1,0)) );
	vector< vector< vector<float> > > cc_xy(nlay+1, vector< vector<float> > (nrow+1, vector<float>(ncol+1,0)) );
	vector< vector< vector<float> > > cc_xz(nlay+1, vector< vector<float> > (nrow+1, vector<float>(ncol+1,0)) );
	vector< vector< vector<float> > > cr_yx(nlay+1, vector< vector<float> > (nrow+1, vector<float>(ncol+1,0)) );
	vector< vector< vector<float> > > cr_yy(nlay+1, vector< vector<float> > (nrow+1, vector<float>(ncol+1,0)) );
	vector< vector< vector<float> > > cr_yz(nlay+1, vector< vector<float> > (nrow+1, vector<float>(ncol+1,0)) );	
	vector< vector< vector<float> > > cv_zx(nlay+1, vector< vector<float> > (nrow+1, vector<float>(ncol+1,0)) );
	vector< vector< vector<float> > > cv_zy(nlay+1, vector< vector<float> > (nrow+1, vector<float>(ncol+1,0)) );
	vector< vector< vector<float> > > cv_zz(nlay+1, vector< vector<float> > (nrow+1, vector<float>(ncol+1,0)) );
	
	vector< vector< vector<float> > > A(nlay+1, vector< vector<float> > (nrow+1, vector<float>(ncol+1,0)) );
	vector< vector< vector<float> > > B(nlay+1, vector< vector<float> > (nrow+1, vector<float>(ncol+1,0)) );
	vector< vector< vector<float> > > C(nlay+1, vector< vector<float> > (nrow+1, vector<float>(ncol+1,0)) );
	vector< vector< vector<float> > > D(nlay+1, vector< vector<float> > (nrow+1, vector<float>(ncol+1,0)) );
	vector< vector< vector<float> > > E(nlay+1, vector< vector<float> > (nrow+1, vector<float>(ncol+1,0)) );
	vector< vector< vector<float> > > F(nlay+1, vector< vector<float> > (nrow+1, vector<float>(ncol+1,0)) );
	vector< vector< vector<float> > > G(nlay+1, vector< vector<float> > (nrow+1, vector<float>(ncol+1,0)) );
	vector< vector< vector<float> > > H(nlay+1, vector< vector<float> > (nrow+1, vector<float>(ncol+1,0)) );
	vector< vector< vector<float> > > I(nlay+1, vector< vector<float> > (nrow+1, vector<float>(ncol+1,0)) );
	vector< vector< vector<float> > > J(nlay+1, vector< vector<float> > (nrow+1, vector<float>(ncol+1,0)) );
	vector< vector< vector<float> > > K(nlay+1, vector< vector<float> > (nrow+1, vector<float>(ncol+1,0)) );
	vector< vector< vector<float> > > L(nlay+1, vector< vector<float> > (nrow+1, vector<float>(ncol+1,0)) );
	vector< vector< vector<float> > > M(nlay+1, vector< vector<float> > (nrow+1, vector<float>(ncol+1,0)) );
	vector< vector< vector<float> > > N(nlay+1, vector< vector<float> > (nrow+1, vector<float>(ncol+1,0)) );
	vector< vector< vector<float> > > O(nlay+1, vector< vector<float> > (nrow+1, vector<float>(ncol+1,0)) );
	vector< vector< vector<float> > > P(nlay+1, vector< vector<float> > (nrow+1, vector<float>(ncol+1,0)) );
	vector< vector< vector<float> > > Q(nlay+1, vector< vector<float> > (nrow+1, vector<float>(ncol+1,0)) );
	vector< vector< vector<float> > > R(nlay+1, vector< vector<float> > (nrow+1, vector<float>(ncol+1,0)) );
	vector< vector< vector<float> > > S(nlay+1, vector< vector<float> > (nrow+1, vector<float>(ncol+1,0)) );
	vector< vector< vector<float> > > RHS(nlay+1, vector< vector<float> > (nrow+1, vector<float>(ncol+1,0)) );

	vector< vector< vector<float> > > head(nlay+2, vector< vector<float> > (nrow+2, vector<float>(ncol+2,0)) );

	vector< vector< vector<float> > > qxx(nlay+1, vector< vector<float> > (nrow+1, vector<float>(ncol+1,0)) );
	vector< vector< vector<float> > > qyy(nlay+1, vector< vector<float> > (nrow+1, vector<float>(ncol+1,0)) );
	vector< vector< vector<float> > > qzz(nlay+1, vector< vector<float> > (nrow+1, vector<float>(ncol+1,0)) );
	vector< vector< vector<float> > > q_well(nlay+1, vector< vector<float> > (nrow+1, vector<float>(ncol+1,0)) );

	vector<float>well_x(n_well+1,0);
	vector<float>well_y(n_well+1,0);
	vector<float>well_z(n_well+1,0);
	vector<float>well_q(n_well+1,0);

	Read_data(ncol,nrow,nlay,kxx_file,kyy_file,kzz_file,ihead_file,ibound_file,dx_3d,dy_3d,dz_3d,head,ibo,cc_xx,cc_xy,cc_xz,
			  cr_yx,cr_yy,cr_yz,cv_zx, cv_zy,cv_zz,well_file,n_well,well_x,well_y,well_z,well_q,dx_file,dy_file,dz_file,
			  f_dx,f_dy,f_dz,flag_dx,flag_dy,flag_dz,debug);

	Set_boundary(ncol,nrow,nlay,kxx_file,kyy_file,kzz_file,ibo,cc_xx,cc_xy,cc_xz,cr_yx,cr_yy,cr_yz,cv_zx, cv_zy,cv_zz,debug);
	
	if (n_well>0)
		modifyK_well(ncol,nrow,nlay,kxx_file,kyy_file,kzz_file,cc_xx,cc_xy,cc_xz,cr_yx,cr_yy,cr_yz,cv_zx, cv_zy,cv_zz,well_x,well_y,well_z,n_well,debug);

	Compute_abcs(nlay, nrow, ncol, dx_3d, dy_3d, dz_3d,cc_xx, cc_xy,cc_xz,cr_yx,cr_yy,cr_yz,cv_zx,cv_zy,cv_zz,A,B,C,D,E,F,G,
				H,I,J,K,L,M,N,O,P,Q,R,S,RHS,ibo,debug);

	if (n_well>0)
		modify_RHS_well(nlay, nrow, ncol,well_x,well_y,well_z,well_q,n_well,RHS);
	
	if(Iterate(max_iter, nlay,nrow, ncol, debug, max_head_change, closure, accl,head,A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,
		 RHS,ibo,CONVERGENCE,NO_CONVERGENCE)==CONVERGENCE)
	{
		Print_head(nlay,nrow, ncol,head_file,head,ibo);
		cout<<"I am working. Please wait.....";
		Compute_flow(nlay,nrow,ncol,dx_3d,dy_3d,dz_3d,head,cc_xx,cc_xy,cc_xz,cr_yx,cr_yy,cr_yz,cv_zx,cv_zy,cv_zz,
						qxx,qyy,qzz,debug);
		Print_flow(nlay,nrow,ncol, qx_file,average_qx,qxx);
		Print_flow(nlay,nrow,ncol, qy_file,average_qy,qyy);
		Print_flow(nlay,nrow,ncol, qz_file,average_qz,qzz);
	}
	else 
		cout<<"Max Iteration achieved without convergence!!\n";

	if (debug>=1)
		Check_flow(nlay, nrow, ncol,qxx,qyy,qzz,RHS,head,dx_3d,dy_3d,dz_3d);	

}

//--------------------------------------------------------------------------------------------------------//
//           Read parameters for block size, convergence, file names                                       //
//--------------------------------------------------------------------------------------------------------//
void Read_para(int &ncol,int &nrow,int &nlay,float &f_dx,float &f_dy,float &f_dz,int &max_iter,float &accl,float &closure,
			   string &kxx_file,string &kyy_file,string &kzz_file,string &ihead_file,string &ibound_file,
			   string &head_file,string &qx_file,string &qy_file,string &qz_file,int &n_well,string &well_file,
			   string &dx_file,string &dy_file,string &dz_file,int &flag_dx,int &flag_dy,int &flag_dz,int &debug)
{
//------read size of block, convergence standard, and output filename-----------
	cout<<"************************************"<<endl;
	cout<<"* Flow simulation with full tensor *"<<endl;
	cout<<"*        U.P.V ( Oct.2009)         *"<<endl;
	cout<<"************************************"<<endl;
	cout<<"                                    "<<endl;
	//cout<<"Parameter file:";
//	string filename;
//	cin>>filename;
	ifstream paraFile;	
	paraFile.open("flowxyz.par");

	string comments;
	int i;
	for (i=1;i<=6;i++)
		getline(paraFile,comments,'\n');

	paraFile>>ncol>>nrow>>nlay;
	getline(paraFile,comments,'\n');
	paraFile>>flag_dx>>flag_dy>>flag_dz;
	getline(paraFile,comments,'\n');
	paraFile>>f_dx>>f_dy>>f_dz;
	getline(paraFile,comments,'\n');
	paraFile>>dx_file;
	getline(paraFile,comments,'\n');
	paraFile>>dy_file;
	getline(paraFile,comments,'\n');
	paraFile>>dz_file;
	getline(paraFile,comments,'\n');
	paraFile>>accl>>closure>>max_iter;
	getline(paraFile,comments,'\n');
	paraFile>>kxx_file;
	getline(paraFile,comments,'\n');
	paraFile>>kyy_file;
	getline(paraFile,comments,'\n');
	paraFile>>kzz_file;
	getline(paraFile,comments,'\n');
	paraFile>>ihead_file;
	getline(paraFile,comments,'\n');
	paraFile>>ibound_file;
	getline(paraFile,comments,'\n');
	paraFile>>head_file;
	getline(paraFile,comments,'\n');
	paraFile>>qx_file;
	getline(paraFile,comments,'\n');
	paraFile>>qy_file;
	getline(paraFile,comments,'\n');
	paraFile>>qz_file;
	getline(paraFile,comments,'\n');
	//-----read well information
	getline(paraFile,comments,'\n');
	paraFile>>n_well;
	getline(paraFile,comments,'\n');
	paraFile>>well_file;
	//-----debug
	getline(paraFile,comments,'\n');
	getline(paraFile,comments,'\n');
	paraFile>>debug;
	getline(paraFile,comments,'\n');
	paraFile.close();


	if (debug>=1)
	{
		ofstream debug_file;
		debug_file.open("debug.dbg");
		debug_file<<"ncol,nrow,nlay---->"<<ncol<<"  "<<nrow<<"  "<<nlay <<endl;
		debug_file<<"f_dx,   f_dy,   f_dz---->"<<f_dx<<" "<<f_dy<<" "<<f_dz<<endl;
		debug_file<<"accl, closure,max_iter--->"<<accl<<" "<<closure<<" "<<max_iter<<endl;
		debug_file<<"kxx file---->"<<kxx_file<<endl;
		debug_file<<"kyy file---->"<<kyy_file<<endl;
		debug_file<<"kzz file---->"<<kzz_file<<endl;
		debug_file<<"ihead file---->"<<ihead_file<<endl;
		debug_file<<"ibound file---->"<<ibound_file<<endl;
		debug_file<<"head file---->"<<head_file<<endl;
		debug_file<<"qx file---->"<<qx_file<<endl;
		debug_file<<"qy file---->"<<qy_file<<endl;
		debug_file<<"qz file---->"<<qz_file<<endl;
		debug_file<<"number of well---->"<<n_well<<endl;
		debug_file<<"well file---->"<<well_file<<endl;
		debug_file.close();
	}
}
//-----------------------------------------------------------------------------------------------------------//
//                          Read conductance,initial head and boundary condition                                                          //
//-----------------------------------------------------------------------------------------------------------//
void Read_data(int ncol,int nrow,int nlay,string &kxx_file,string &kyy_file,string &kzz_file,
			   string &ihead_file,string &ibound_file,vector< vector< vector<float> > > &dx_3d,
			   vector< vector< vector<float> > > &dy_3d,vector< vector< vector<float> > > &dz_3d,
			   vector< vector< vector<float> > > &head,
			   vector< vector< vector<int> > > &ibo,vector< vector< vector<float> > > &cc_xx,
			   vector< vector< vector<float> > > &cc_xy,vector< vector< vector<float> > > &cc_xz,
			   vector< vector< vector<float> > > &cr_yx,vector< vector< vector<float> > > &cr_yy,
			   vector< vector< vector<float> > > &cr_yz,vector< vector< vector<float> > > &cv_zx,
			   vector< vector< vector<float> > > &cv_zy,vector< vector< vector<float> > > &cv_zz,
			   string &well_file,int n_well,vector<float>&well_x,vector<float>&well_y,
			   vector<float>&well_z,vector<float>&well_q,string &dx_file,string &dy_file,string &dz_file,
			   float f_dx,float f_dy,float f_dz,int flag_dx,int flag_dy,int flag_dz,int debug)
{	
	int i,j,k;
	string comments;
//------------read the dx
	if (flag_dx==1){
	ifstream dxFile;
	dxFile.open(dx_file.c_str());
	for (i=1;i<=3;i++)
		getline(dxFile,comments,'\n');

	for (j=1;j<=ncol;j++)
		dxFile>>dx_3d[0][0][j];	
	dx_3d[0][0][0]=dx_3d[0][0][1];dx_3d[0][0][ncol+1]=dx_3d[0][0][ncol];
	for (i=1;i<=nrow+1;i++)
		for (j=0;j<=ncol+1;j++)
			dx_3d[0][i][j]=dx_3d[0][0][j];
	for (k=1;k<=nlay+1;k++)
		for (i=0;i<=nrow+1;i++)
			for (j=0;j<=ncol+1;j++)
				dx_3d[k][i][j]=dx_3d[0][i][j];
	}
	else
	{
	for (k=0;k<=nlay+1;k++)
		for (i=0;i<=nrow+1;i++)
			for (j=0; j<=ncol+1;j++)
				dx_3d[k][i][j]=f_dx;
	}

//------------read the dy
	if (flag_dy==1){
	ifstream dyFile;
	dyFile.open(dy_file.c_str());
	for (i=1;i<=3;i++)
		getline(dyFile,comments,'\n');

	for (i=nrow;i>=1;i--)
		dyFile>>dy_3d[0][i][0];	
	dy_3d[0][0][0]=dy_3d[0][1][0];dy_3d[0][nrow+1][0]=dy_3d[0][nrow][0];
	for (i=0;i<=nrow+1;i++)
		for (j=0;j<=ncol+1;j++)
			dy_3d[0][i][j]=dy_3d[0][i][0];
	for (k=1;k<=nlay+1;k++)
		for (i=0;i<=nrow+1;i++)
			for (j=0; j<=ncol+1;j++)
				dy_3d[k][i][j]=dy_3d[0][i][j];
	}
	else
	{
	for (k=0;k<=nlay+1;k++)
		for (i=0;i<=nrow+1;i++)
			for (j=0; j<=ncol+1;j++)
				dy_3d[k][i][j]=f_dy;
	}
//------------read the dz
	if (flag_dz==1){
	ifstream dzFile;
	dzFile.open(dz_file.c_str());
	for (i=1;i<=3;i++)
		getline(dzFile,comments,'\n');

	for (k=nlay;k>=1;k--)
		dzFile>>dz_3d[k][0][0];	
	dy_3d[0][0][0]=dy_3d[1][0][0];dy_3d[nlay+1][0][0]=dy_3d[nlay][0][0];
	for (k=0;k<=nlay+1;k++)
		for (i=0;i<=nrow+1;i++)
			for (j=0; j<=ncol+1;j++)
				dz_3d[k][i][j]=dz_3d[k][0][0];
	}
	else
	{
	for (k=0;k<=nlay+1;k++)
		for (i=0;i<=nrow+1;i++)
			for (j=0; j<=ncol+1;j++)
				dz_3d[k][i][j]=f_dz;
	}
//---------read conductances
	ifstream condxFile;
	condxFile.open(kxx_file.c_str());
	for (i=1;i<=5;i++)
		getline(condxFile,comments,'\n');

	for (k=nlay;k>=1;k--)
		for (i=nrow;i>=1;i--)
			for (j=1; j<=ncol-1;j++)
				condxFile>>cc_xx[k][i][j]>>cc_xy[k][i][j]>>cc_xz[k][i][j];
	condxFile.close();

	ifstream condyFile;
	condyFile.open(kyy_file.c_str());
	for (i=1;i<=5;i++)
		getline(condyFile,comments,'\n');
	for (k=nlay;k>=1;k--)
		for (i=nrow-1;i>=1;i--)
			for (j=1; j<=ncol;j++)
				condyFile>>cr_yx[k][i][j]>>cr_yy[k][i][j]>>cr_yz[k][i][j];
	condyFile.close();

	ifstream condzFile;
	condzFile.open(kzz_file.c_str());
	for (i=1;i<=5;i++)
		getline(condzFile,comments,'\n');
	for (k=nlay-1;k>=1;k--)
		for (i=nrow;i>=1;i--)
			for (j=1; j<=ncol;j++)
				condzFile>>cv_zx[k][i][j]>>cv_zy[k][i][j]>>cv_zz[k][i][j];
	condzFile.close();

//-----------read intial head
	ifstream ihFile;
	ihFile.open(ihead_file.c_str());
	for (i=1;i<=3;i++)
		getline(ihFile,comments,'\n');
	for (k=nlay;k>=1;k--)
		for (i=nrow;i>=1;i--)
			for (j=1; j<=ncol;j++)
				ihFile>>head[k][i][j];
	ihFile.close();
//-----------read boundary condition
	ifstream ibFile;
	ibFile.open(ibound_file.c_str());
	for (i=1;i<=3;i++)
		getline(ibFile,comments,'\n');
	for (k=nlay;k>=1;k--)
		for (i=nrow;i>=1;i--)
			for (j=1; j<=ncol;j++)
				ibFile>>ibo[k][i][j];
	ibFile.close();
//-----------read well information
	ifstream wellFile;
	wellFile.open(well_file.c_str());
	getline(wellFile,comments,'\n');
	for (int iw=1;iw<=n_well;iw++){
		wellFile>>well_x[iw]>>well_y[iw]>>well_z[iw]>>well_q[iw];
		getline(wellFile,comments,'\n');
	}
//-----------DEBUG
	if (debug>=1)
	{
		write_debug1(nlay,nrow,ncol,ibound_file,ibo);
		write_debug2(nlay,nrow,ncol,ihead_file,head);
		write_debug3(nlay,nrow,ncol,kxx_file,cc_xx);
		write_debug3(nlay,nrow,ncol,kxx_file,cc_xy);
		write_debug3(nlay,nrow,ncol,kxx_file,cc_xz);
		write_debug3(nlay,nrow,ncol,kyy_file,cr_yx);
		write_debug3(nlay,nrow,ncol,kyy_file,cr_yy);
		write_debug3(nlay,nrow,ncol,kyy_file,cr_yz);
		write_debug3(nlay,nrow,ncol,kzz_file,cv_zx);
		write_debug3(nlay,nrow,ncol,kzz_file,cv_zy);
		write_debug3(nlay,nrow,ncol,kzz_file,cv_zz);
	}

}
//-----------------------------------------------------------------//
//                 modify k for no flow boundary condition         //
//-----------------------------------------------------------------//
void Set_boundary(int ncol,int nrow,int nlay,
				string &kxx_file,string &kyy_file,string &kzz_file,vector< vector< vector<int> > > &ibo,
			   vector< vector< vector<float> > > &cc_xx,vector< vector< vector<float> > > &cc_xy,
			   vector< vector< vector<float> > > &cc_xz,vector< vector< vector<float> > > &cr_yx,
			   vector< vector< vector<float> > > &cr_yy,vector< vector< vector<float> > > &cr_yz,
			   vector< vector< vector<float> > > &cv_zx,vector< vector< vector<float> > > &cv_zy,
			   vector< vector< vector<float> > > &cv_zz,int debug)
{
	int i,j,k;
	for (k=1;k<=nlay;k++)
		for (i=1;i<=nrow;i++)
			for (j=1; j<=ncol;j++)
				if (ibo[k][i][j]==0)
				{
					cc_xx[k][i][j]=0; cc_xx[k][i][j-1]=0;
					cc_xy[k][i][j]=0; cc_xy[k][i][j-1]=0;
					cc_xz[k][i][j]=0; cc_xz[k][i][j-1]=0;

					cr_yx[k][i][j]=0; cr_yx[k][i-1][j]=0;
					cr_yy[k][i][j]=0; cr_yy[k][i-1][j]=0;
					cr_yz[k][i][j]=0; cr_yz[k][i-1][j]=0;

					cv_zx[k][i][j]=0; cv_zx[k-1][i][j]=0;
					cv_zy[k][i][j]=0; cv_zy[k-1][i][j]=0;
					cv_zz[k][i][j]=0; cv_zz[k-1][i][j]=0;
				}
	if (debug>=1)
	{
	//	cout<<"The hydraulic conductances after setting the boundary conditions"<<endl;
		write_debug3(nlay,nrow,ncol,kxx_file,cc_xx);
		write_debug3(nlay,nrow,ncol,kxx_file,cc_xy);
		write_debug3(nlay,nrow,ncol,kxx_file,cc_xz);
		write_debug3(nlay,nrow,ncol,kyy_file,cr_yx);
		write_debug3(nlay,nrow,ncol,kyy_file,cr_yy);
		write_debug3(nlay,nrow,ncol,kyy_file,cr_yz);
		write_debug3(nlay,nrow,ncol,kzz_file,cv_zx);
		write_debug3(nlay,nrow,ncol,kzz_file,cv_zy);
		write_debug3(nlay,nrow,ncol,kzz_file,cv_zz);
	}


}
//-----------------------------------------------------------------//
//                 write to debug (type int)                       //
//-----------------------------------------------------------------//
void write_debug1(int nlay,int nrow,int ncol,string &ibound_file,vector< vector< vector<int> > > &ibo)
{
	int i,j,k;
	ofstream debug_file;
	debug_file.open("debug.dbg",ios::out|ios::app);
	for (k=0;k<=nlay;k++)
	{
		debug_file<<ibound_file<<"layer"<<k<<endl;
		for (i=0;i<=nrow;i++)
		{
			for (j=0;j<=ncol;j++)
			{
				debug_file.setf(ios_base::right,ios_base::adjustfield);
				debug_file.precision(8);
				debug_file.setf(ios_base::showpoint);
				ios_base::fmtflags old=debug_file.setf(ios_base::scientific,ios_base::floatfield);
				debug_file.width(4);
				debug_file<<ibo[k][i][j];
			}
			debug_file<<endl;
		}
		debug_file<<endl;
	}
	debug_file.close();
}
//-----------------------------------------------------------------//
//                 write to debug (head)                     //
//-----------------------------------------------------------------//
void write_debug2(int nlay,int nrow,int ncol,string &ihead_file,vector< vector< vector<float> > > &head)
{
	int i,j,k;
	ofstream debug_file;
	debug_file.open("debug.dbg",ios::out|ios::app);
	for (k=0;k<=nlay+1;k++)
	{
		debug_file<<ihead_file<<"layer"<<k<<endl;
		for (i=0;i<=nrow+1;i++)
		{
			for (j=0;j<=ncol+1;j++)
			{
				debug_file.setf(ios_base::right,ios_base::adjustfield);
				debug_file.precision(8);
				debug_file.setf(ios_base::showpoint);
				ios_base::fmtflags old=debug_file.setf(ios_base::scientific,ios_base::floatfield);
				debug_file.width(20);
				debug_file<<head[k][i][j];
			}
			debug_file<<endl;
		}
		debug_file<<endl;
	}
	debug_file.close();
}
//-----------------------------------------------------------------//
//                 write to debug (cc,cr,cv)                     //
//-----------------------------------------------------------------//
void write_debug3(int nlay,int nrow,int ncol,string &kxx_file,vector< vector< vector<float> > > &cc_xy)
{
	int i,j,k;
	ofstream debug_file;
	debug_file.open("debug.dbg",ios::out|ios::app);
	for (k=0;k<=nlay;k++)
	{
		debug_file<<kxx_file<<"  "<<"cc_cr_cv"<<"  "<<"layer"<<k<<endl;
		for (i=0;i<=nrow;i++)
		{
			for (j=0;j<=ncol;j++)
			{
				debug_file.setf(ios_base::right,ios_base::adjustfield);
				debug_file.precision(8);
				debug_file.setf(ios_base::showpoint);
				ios_base::fmtflags old=debug_file.setf(ios_base::scientific,ios_base::floatfield);
				debug_file.width(20);
				debug_file<<cc_xy[k][i][j];
			}
			debug_file<<endl;
		}
		debug_file<<endl;
	}
	debug_file.close();
}

//-----------------------------------------------------------------//
//                 compute coefficient matrix A--S                 //
//-----------------------------------------------------------------//
void Compute_abcs(int nlay,int nrow,int ncol,vector< vector< vector<float> > >dx,vector< vector< vector<float> > >dy,
			   vector< vector< vector<float> > >dz,vector< vector< vector<float> > > &cc_xx,
			   vector< vector< vector<float> > > &cc_xy,vector< vector< vector<float> > > &cc_xz,
			   vector< vector< vector<float> > > &cr_yx,vector< vector< vector<float> > > &cr_yy,
			   vector< vector< vector<float> > > &cr_yz,vector< vector< vector<float> > > &cv_zx,
			   vector< vector< vector<float> > > &cv_zy,vector< vector< vector<float> > > &cv_zz,
			   vector< vector< vector<float> > > &A,vector< vector< vector<float> > > &B,
			   vector< vector< vector<float> > > &C,vector< vector< vector<float> > > &D,
			   vector< vector< vector<float> > > &E,vector< vector< vector<float> > > &F,
			   vector< vector< vector<float> > > &G,vector< vector< vector<float> > > &H,
			   vector< vector< vector<float> > > &I,vector< vector< vector<float> > > &J,
			   vector< vector< vector<float> > > &K,vector< vector< vector<float> > > &L,
			   vector< vector< vector<float> > > &M,vector< vector< vector<float> > > &N,
			   vector< vector< vector<float> > > &O,vector< vector< vector<float> > > &P,
			   vector< vector< vector<float> > > &Q,vector< vector< vector<float> > > &R,
			   vector< vector< vector<float> > > &S,vector< vector< vector<float> > > &RHS,
			   vector< vector< vector<int> > > ibo,int debug)
{
	int i,j,k;
	float c1,c2,c3,c4,c5,c6,c7,c8,c9,c10,c11,c12,c13,c14,c15,
		  c16,c17,c18,c19,c20,c21,c22,c23,c24,c25,c26,c27,c28,c29,c30;
	for (k=1;k<=nlay;k++)
		for (i=1;i<=nrow;i++)
			for (j=1;j<=ncol;j++)
			{
				c1=2*  cc_xx[k][i][j]/(dx[k][i][j]+dx[k][i][j+1]);
				c2=0.5*cc_xy[k][i][j]/(0.5*dy[k][i+1][j+1]+0.5*dy[k][i-1][j+1]+dy[k][i][j+1]);
				c3=0.5*cc_xy[k][i][j]/(0.5*dy[k][i+1][j]+0.5*dy[k][i-1][j]+dy[k][i][j]);
				c4=0.5*cc_xz[k][i][j]/(0.5*dz[k+1][i][j+1]+0.5*dz[k-1][i][j+1]+dz[k][i][j+1]);
				c5=0.5*cc_xz[k][i][j]/(0.5*dz[k+1][i][j]+0.5*dz[k-1][i][j]+dz[k][i][j]);
				c6=-2*cc_xx[k][i][j-1]/(    dx[k][i][j]+dx[k][i][j-1]);
				c7=-0.5*cc_xy[k][i][j-1]/(0.5*dy[k][i+1][j]+0.5*dy[k][i-1][j]+dy[k][i][j]);
				c8=-0.5*cc_xy[k][i][j-1]/(0.5*dy[k][i+1][j-1]+0.5*dy[k][i-1][j-1]+dy[k][i][j-1]);
				c9=-0.5*cc_xz[k][i][j-1]/(0.5*dz[k+1][i][j]+0.5*dz[k-1][i][j]+dy[k][i][j]);
				c10=-0.5*cc_xz[k][i][j-1]/(0.5*dz[k+1][i][j-1]+0.5*dz[k-1][i][j-1]+dz[k][i][j-1]);

				c11=0.5*cr_yx[k][i][j]/(0.5*dx[k][i][j+1]+0.5*dx[k][i][j-1]+dx[k][i][j]);
				c12=0.5*cr_yx[k][i][j]/(0.5*dx[k][i+1][j+1]+0.5*dx[k][i+1][j-1]+dx[k][i+1][j]);
				c13=2* cr_yy[k][i][j]/(dy[k][i][j]+dy[k][i+1][j]);
				c14=0.5*cr_yz[k][i][j]/(0.5*dz[k+1][i][j]+0.5*dz[k-1][i][j]+dz[k][i][j]);
				c15=0.5*cr_yz[k][i][j]/(0.5*dz[k+1][i+1][j]+0.5*dz[k-1][i+1][j]+dz[k][i+1][j]);
				c16=-0.5*cr_yx[k][i-1][j]/(0.5*dx[k][i][j+1]+0.5*dx[k][i][j-1]+dx[k][i][j]);
				c17=-0.5*cr_yx[k][i-1][j]/(0.5*dx[k][i-1][j+1]+0.5*dx[k][i-1][j-1]+dx[k][i-1][j]);
				c18=-2*cr_yy[k][i-1][j]/(dy[k][i][j]+dy[k][i-1][j]);
				c19=-0.5*cr_yz[k][i-1][j]/(0.5*dz[k+1][i][j]+0.5*dz[k-1][i][j]+dz[k][i][j]);
				c20=-0.5*cr_yz[k][i-1][j]/(0.5*dz[k+1][i-1][j]+0.5*dz[k-1][i-1][j]+dz[k][i-1][j]);

				c21=0.5*cv_zx[k][i][j]/(0.5*dx[k][i][j+1]+0.5*dx[k][i][j-1]+dx[k][i][j]);
				c22=0.5*cv_zx[k][i][j]/(0.5*dx[k+1][i][j+1]+0.5*dx[k+1][i][j-1]+dx[k+1][i][j]);
				c23=0.5*cv_zy[k][i][j]/(0.5*dy[k][i+1][j]+0.5*dy[k][i-1][j]+dy[k][i][j]);
				c24=0.5*cv_zy[k][i][j]/(0.5*dy[k+1][i+1][j]+0.5*dy[k+1][i-1][j]+dy[k+1][i][j]);
				c25=2*cv_zz[k][i][j]/(dz[k+1][i][j]+dz[k][i][j]);
				c26=-0.5*cv_zx[k-1][i][j]/(0.5*dx[k][i][j+1]+0.5*dx[k][i][j-1]+dx[k][i][j]);
				c27=-0.5*cv_zx[k-1][i][j]/(0.5*dx[k-1][i][j+1]+0.5*dx[k-1][i][j-1]+dx[k-1][i][j]);
				c28=-0.5*cv_zy[k-1][i][j]/(0.5*dy[k][i+1][j]+0.5*dy[k][i-1][j]+dy[k][i][j]);
				c29=-0.5*cv_zy[k-1][i][j]/(0.5*dy[k-1][i+1][j]+0.5*dy[k-1][i-1][j]+dy[k-1][i][j]);
				c30=-2*cv_zz[k-1][i][j]/(dz[k-1][i][j]+dz[k][i][j]);


				A[k][i][j]=dy[k][i][j]*dz[k][i][j]*c1+
					       dx[k][i][j]*dz[k][i][j]*(c11+c16)+
						   dx[k][i][j]*dy[k][i][j]*(c21+c26);

				B[k][i][j]=dy[k][i][j]*dz[k][i][j]*(-c1+c6)+
					       dx[k][i][j]*dz[k][i][j]*(-c13+c18)+
						   dx[k][i][j]*dy[k][i][j]*(-c25+c30);

				C[k][i][j]=dy[k][i][j]*dz[k][i][j]*c2+
					       dx[k][i][j]*dz[k][i][j]*c12;
	
				D[k][i][j]=dy[k][i][j]*dz[k][i][j]*(-c2)+
					       dx[k][i][j]*dz[k][i][j]*c17;

				E[k][i][j]=dy[k][i][j]*dz[k][i][j]*(c3+c7)+
					       dx[k][i][j]*dz[k][i][j]*c13+
						   dx[k][i][j]*dy[k][i][j]*(c23+c28);
	
				F[k][i][j]=dy[k][i][j]*dz[k][i][j]*(-c3-c7)+
					       dx[k][i][j]*dz[k][i][j]*(-c18)+
						   dx[k][i][j]*dy[k][i][j]*(-c23-c28);
	
				G[k][i][j]=dy[k][i][j]*dz[k][i][j]*c4+
						   dx[k][i][j]*dy[k][i][j]*c22;
			
				H[k][i][j]=dy[k][i][j]*dz[k][i][j]*(-c4)+
						   dx[k][i][j]*dy[k][i][j]*c27;
		
				I[k][i][j]=dy[k][i][j]*dz[k][i][j]*(c5+c9)+
					       dx[k][i][j]*dz[k][i][j]*(c14+c19)+
						   dx[k][i][j]*dy[k][i][j]*c25;
	
				J[k][i][j]=dy[k][i][j]*dz[k][i][j]*(-c5-c9)+
					       dx[k][i][j]*dz[k][i][j]*(-c14-c19)+
						   dx[k][i][j]*dy[k][i][j]*(-c30);
			
				K[k][i][j]=dy[k][i][j]*dz[k][i][j]*(-c6)+
					       dx[k][i][j]*dz[k][i][j]*(-c11-c16)+
						   dy[k][i][j]*dx[k][i][j]*(-c21-c26);
		
				L[k][i][j]=dy[k][i][j]*dz[k][i][j]*(c8)+
					       dx[k][i][j]*dz[k][i][j]*(-c12);

				M[k][i][j]=dy[k][i][j]*dz[k][i][j]*(-c8)+
					       dx[k][i][j]*dz[k][i][j]*(-c17);
						 
				N[k][i][j]=dy[k][i][j]*dz[k][i][j]*(c10)+
						   dy[k][i][j]*dx[k][i][j]*(-c22);
			
				O[k][i][j]=dy[k][i][j]*dz[k][i][j]*(-c10)+
						   dy[k][i][j]*dx[k][i][j]*(-c27);

				P[k][i][j]=dx[k][i][j]*dz[k][i][j]*(c15)+
						   dy[k][i][j]*dx[k][i][j]*(c24);
		
				Q[k][i][j]=dx[k][i][j]*dz[k][i][j]*(-c15)+
						   dy[k][i][j]*dx[k][i][j]*(c29);
			
				R[k][i][j]=dx[k][i][j]*dz[k][i][j]*(c20)+
						   dy[k][i][j]*dx[k][i][j]*(-c24);
		
				S[k][i][j]=dx[k][i][j]*dz[k][i][j]*(-c20)+
						   dy[k][i][j]*dx[k][i][j]*(-c29);
			}

	if(debug>=1)
	{	
		ofstream debug_file;
		debug_file.open("debug.dbg",ios::out|ios::app);
		debug_file<<"A"<<"  "<<"B"<<"  "<<"C"<<"  "<<"D"<<"  "<<"E"<<"  "<<"F"<<"  "<<"G"<<"  "<<"H"<<"  "<<"I"<<"  "
					<<"J"<<"  "<<"K"<<"  "<<"L"<<"  "<<"M"<<"  "<<"N"<<"  "<<"O"<<"  "<<"P"<<"  "<<"Q"<<"  "<<"R"<<"  "
					<<"S"<<"  "<<"RHS"<<endl;
		for (k=1;k<=nlay;k++)
		{
			debug_file<<"layer"<<"  "<<k<<endl;
			for (i=1;i<=nrow;i++)
			{
				for (j=1;j<=ncol;j++)
				{
					debug_file.setf(ios_base::right,ios_base::adjustfield);
					debug_file.precision(8);
					debug_file.setf(ios_base::showpoint);
					ios_base::fmtflags old=debug_file.setf(ios_base::scientific,ios_base::floatfield);
					debug_file.width(20);
					debug_file<<A[k][i][j]<<"   "<<B[k][i][j]<<"   "<<C[k][i][j]<<"   "<<D[k][i][j]<<"   "
						<<E[k][i][j]<<"   "<<F[k][i][j]<<"   "<<G[k][i][j]<<"   "<<H[k][i][j]<<"  "
						<<I[k][i][j]<<"   "<<J[k][i][j]<<"   "<<K[k][i][j]<<"   "<<L[k][i][j]<<"   "
						<<M[k][i][j]<<"   "<<N[k][i][j]<<"   "<<O[k][i][j]<<"   "<<P[k][i][j]<<"   "
						<<Q[k][i][j]<<"   "<<R[k][i][j]<<"   "<<S[k][i][j]<<"    "<<RHS[k][i][j]<<endl;
				}
				debug_file<<endl;
			}
			debug_file<<endl;
		}

		debug_file.close();
	}

}
//---------------------------------------------------------------------------//
//                        SOR iteration                                      //
//---------------------------------------------------------------------------//
bool Iterate(int max_iter,int nlay,int nrow,int ncol,int debug,float max_head_change,float closure,float accl,
			vector< vector< vector<float> > > &head,vector< vector< vector<float> > > &A,
			vector< vector< vector<float> > > &B,vector< vector< vector<float> > > &C,
			vector< vector< vector<float> > > &D,vector< vector< vector<float> > > &E,
			vector< vector< vector<float> > > &F, vector< vector< vector<float> > > &G,
			vector< vector< vector<float> > > &H,vector< vector< vector<float> > > &I,
			vector< vector< vector<float> > > &J, vector< vector< vector<float> > > &K,
			vector< vector< vector<float> > > &L, vector< vector< vector<float> > > &M,
			vector< vector< vector<float> > > &N,vector< vector< vector<float> > > &O,
			vector< vector< vector<float> > > &P,vector< vector< vector<float> > > &Q,
			vector< vector< vector<float> > > &R, vector< vector< vector<float> > > &S,
			vector< vector< vector<float> > > &RHS,
			vector< vector< vector<int> > > ibo,bool CONVERGENCE,bool NO_CONVERGENCE)
{
	int m,ifk,i,j,k;
	ifk=0;
	for(m=1;m<=max_iter;m++)
	{
		Fill_outside_ring(nlay, nrow, ncol,head);

		if(debug>=1 && m==1 && ifk==0)
		{
			ofstream debug_file;
			debug_file.open("debug.dbg",ios::out|ios::app);
			debug_file<<"extended head"<<endl;
			for (k=0;k<=nlay+1;k++)
			{
				debug_file<<"layer"<<"  "<<k<<endl;
				for (i=0;i<=nrow+1;i++)
				{
					for (j=0;j<=ncol+1;j++)
					{
						debug_file.setf(ios_base::right,ios_base::adjustfield);
						debug_file.precision(8);
						debug_file.setf(ios_base::showpoint);
						ios_base::fmtflags old=debug_file.setf(ios_base::scientific,ios_base::floatfield);
						debug_file.width(20);
						debug_file<<head[k][i][j];
					}
					debug_file<<endl;
				}
				debug_file<<endl;
			}

			debug_file.close();
		}//end of if

		ifk=1;//???
		if(One_iteration(nlay,nrow,ncol,max_head_change,closure,accl,head,ibo,A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,RHS,
						debug,CONVERGENCE,NO_CONVERGENCE)==CONVERGENCE) break;			
	}
	if(m>max_iter) 
		return NO_CONVERGENCE;
	else 
		return CONVERGENCE;
}
//---------------------------------------------------------------------------//
//                        extend the head for no flow boundary condition     //
//---------------------------------------------------------------------------//
void Fill_outside_ring(int nlay,int nrow,int ncol,vector< vector< vector<float> > > &head)
{
	int i,j,k;
	for(k=1;k<=nlay;k++)
	{
		head[k][0][0]=head[k][1][1];
		head[k][0][ncol+1]=head[k][1][ncol];
		head[k][nrow+1][0]=head[k][nrow][1];
		head[k][nrow+1][ncol+1]=head[k][nrow][ncol];

		for(i=1;i<=nrow;i++)
		{
			head[k][i][0]=head[k][i][1];
			head[k][i][ncol+1]=head[k][i][ncol];
		}

		for(j=1;j<=ncol;j++)
		{
			head[k][0][j]=head[k][1][j];
			head[k][nrow+1][j]=head[k][nrow][j];
		}
	}

	for(i=0;i<=nrow+1;i++)
		for(j=0;j<=ncol+1;j++)
		{
			head[0][i][j]=head[1][i][j];
			head[nlay+1][i][j]=head[nlay][i][j];
		}
}
//---------------------------------------------------------------------------//
//                       One iteration                                       //
//---------------------------------------------------------------------------//
bool One_iteration(int nlay,int nrow,int ncol,float max_head_change,float closure,float accl,
				  vector< vector< vector<float> > > &head,vector< vector< vector<int> > > ibo,
				  vector< vector< vector<float> > > &A,vector< vector< vector<float> > > &B,
				  vector< vector< vector<float> > > &C,vector< vector< vector<float> > > &D,
				  vector< vector< vector<float> > > &E,vector< vector< vector<float> > > &F,
				  vector< vector< vector<float> > > &G,vector< vector< vector<float> > > &H,
				  vector< vector< vector<float> > > &I,vector< vector< vector<float> > > &J,
				  vector< vector< vector<float> > > &K,vector< vector< vector<float> > > &L,
				  vector< vector< vector<float> > > &M,vector< vector< vector<float> > > &N,
				  vector< vector< vector<float> > > &O,vector< vector< vector<float> > > &P,
				  vector< vector< vector<float> > > &Q,vector< vector< vector<float> > > &R,
				  vector< vector< vector<float> > > &S,vector< vector< vector<float> > > &RHS,
				  int debug,bool CONVERGENCE,bool NO_CONVERGENCE)
{
	int i,j,k;
	int ii,kk,jj;
	float intermediate_h,increment;
	for(k=1;k<=nlay;k++)
		for(i=1;i<=nrow;i++)
    		for (j=1;j<=ncol;j++)
				if(ibo[k][i][j]==1)
				{
				//cope with inactive cells
				for (kk=k-1;kk<=k+1;kk++)
					for (ii=i-1;ii<=i+1;ii++)
						for (jj=j-1;jj<=j+1;jj++)
						{
							if ((kk==k-1 && ii==i-1 && jj==j-1) || (kk==k+1 && ii==i-1 && jj==j-1) ||
								(kk==k-1 && ii==i+1 && jj==j-1) || (kk==k+1 && ii==i+1 && jj==j-1) ||
								(kk==k-1 && ii==i+1 && jj==j+1) || (kk==k+1 && ii==i+1 && jj==j+1) ||
								(kk==k-1 && ii==i-1 && jj==j+1) || (kk==k+1 && ii==i-1 && jj==j+1) ||
								(kk==k   && ii==i   && jj==j  ))  // these cells cannot be used to finite-difference.
								break;
							int count=0;
							float sum_head=0;
							if (ibo[kk][ii][jj]==0)  //the neighbour 6 cells are considered in 3D.
							{
								if (ibo[kk+1][ii][jj]  !=0 ){count=count+1;sum_head=sum_head+head[kk+1][ii  ][jj  ];}
								if (ibo[kk-1][ii][jj]  !=0 ){count=count+1;sum_head=sum_head+head[kk-1][ii  ][jj  ];}
								if (ibo[kk][ii-1][jj]  !=0 ){count=count+1;sum_head=sum_head+head[kk  ][ii-1][jj  ];}
								if (ibo[kk][ii+1][jj]  !=0 ){count=count+1;sum_head=sum_head+head[kk  ][ii+1][jj  ];}
								if (ibo[kk][ii][jj+1]  !=0 ){count=count+1;sum_head=sum_head+head[kk  ][ii  ][jj+1];}
								if (ibo[kk][ii][jj-1]  !=0 ){count=count+1;sum_head=sum_head+head[kk  ][ii  ][jj-1];}

								head[kk][ii][jj]=sum_head/count;									
							}
						}
				//cope with inactive cell
				intermediate_h=(-(A[k][i][j]*head[k][i][j+1]+C[k][i][j]*head[k][i+1][j+1]+D[k][i][j]*head[k][i-1][j+1]
								+E[k][i][j]*head[k][i+1][j]+F[k][i][j]*head[k][i-1][j]+G[k][i][j]*head[k+1][i][j+1]
								+H[k][i][j]*head[k-1][i][j+1]+I[k][i][j]*head[k+1][i][j]+J[k][i][j]*head[k-1][i][j]
								+K[k][i][j]*head[k][i][j-1]+L[k][i][j]*head[k][i+1][j-1]+M[k][i][j]*head[k][i-1][j-1]
								+N[k][i][j]*head[k+1][i][j-1]+O[k][i][j]*head[k-1][i][j-1]+P[k][i][j]*head[k+1][i+1][j]
								+Q[k][i][j]*head[k-1][i+1][j]+R[k][i][j]*head[k+1][i-1][j]+S[k][i][j]*head[k-1][i-1][j])+RHS[k][i][j])
								/B[k][i][j];				
				increment=intermediate_h-head[k][i][j];
				if(fabs(increment)>=max_head_change)
					max_head_change=fabs(increment);
				head[k][i][j]=head[k][i][j]+accl*increment;
				}
	if(max_head_change <=  closure) 
		return (CONVERGENCE);
	else 
		return(NO_CONVERGENCE);
}
//---------------------------------------------------------------------------//
//                       Compute interface flow                              //
//---------------------------------------------------------------------------//
void Compute_flow(int nlay,int nrow,int ncol,vector< vector< vector<float> > >dx,vector< vector< vector<float> > >dy,
				  vector< vector< vector<float> > >dz,vector< vector< vector<float> > > &head,
				  vector< vector< vector<float> > > &cc_xx,vector< vector< vector<float> > > &cc_xy,
				  vector< vector< vector<float> > > &cc_xz,vector< vector< vector<float> > > &cr_yx,
				  vector< vector< vector<float> > > &cr_yy,vector< vector< vector<float> > > &cr_yz,
				  vector< vector< vector<float> > > &cv_zx,vector< vector< vector<float> > > &cv_zy,
				  vector< vector< vector<float> > > &cv_zz,vector< vector< vector<float> > > &qxx,
				  vector< vector< vector<float> > > &qyy,vector< vector< vector<float> > > &qzz,int debug)
{	
	int i,j,k;
	Fill_outside_ring(nlay, nrow, ncol,head);
	if(debug>=1)
		{
			ofstream debug_file;
			debug_file.open("debug.dbg",ios::out|ios::app);
			debug_file<<"finally extended head"<<endl;
			for (k=0;k<=nlay+1;k++)
			{
				debug_file<<"layer"<<"  "<<k<<endl;
				for (i=0;i<=nrow+1;i++)
				{
					for (j=0;j<=ncol+1;j++)
					{
						debug_file.setf(ios_base::right,ios_base::adjustfield);
						debug_file.precision(8);
						debug_file.setf(ios_base::showpoint);
						ios_base::fmtflags old=debug_file.setf(ios_base::scientific,ios_base::floatfield);
						debug_file.width(20);
						debug_file<<head[k][i][j];
					}
					debug_file<<endl;
				}
				debug_file<<endl;
			}

			debug_file.close();
		}
	float c1,c2,c3,c4,c5,c11,c12,c13,c14,c15,
		  c21,c22,c23,c24,c25;
	for(k=1;k<=nlay;k++)
		for(i=1;i<=nrow;i++)
			for (j=1;j<=ncol;j++)
			{
				c1=2*  cc_xx[k][i][j]/(    dx[k][i][j]+dx[k][i][j+1]);
				c2=0.5*cc_xy[k][i][j]/(0.5*dy[k][i+1][j+1]+0.5*dy[k][i-1][j+1]+dy[k][i][j+1]);
				c3=0.5*cc_xy[k][i][j]/(0.5*dy[k][i+1][j]+0.5*dy[k][i-1][j]+dy[k][i][j]);
				c4=0.5*cc_xz[k][i][j]/(0.5*dz[k+1][i][j+1]+0.5*dz[k-1][i][j+1]+dz[k][i][j+1]);
				c5=0.5*cc_xz[k][i][j]/(0.5*dz[k+1][i][j]+0.5*dz[k-1][i][j]+dz[k][i][j]);

				c11=0.5*cr_yx[k][i][j]/(0.5*dx[k][i][j+1]+0.5*dx[k][i][j-1]+dx[k][i][j]);
				c12=0.5*cr_yx[k][i][j]/(0.5*dx[k][i+1][j+1]+0.5*dx[k][i+1][j-1]+dx[k][i+1][j]);
				c13=2*cr_yy[k][i][j]/(dy[k][i][j]+dy[k][i+1][j]);
				c14=0.5*cr_yz[k][i][j]/(0.5*dz[k+1][i][j]+0.5*dz[k-1][i][j]+dz[k][i][j]);
				c15=0.5*cr_yz[k][i][j]/(0.5*dz[k+1][i+1][j]+0.5*dz[k-1][i+1][j]+dz[k][i+1][j]);


				c21=0.5*cv_zx[k][i][j]/(0.5*dx[k][i][j+1]+0.5*dx[k][i][j-1]+dx[k][i][j]);
				c22=0.5*cv_zx[k][i][j]/(0.5*dx[k+1][i][j+1]+0.5*dx[k+1][i][j-1]+dx[k+1][i][j]);
				c23=0.5*cv_zy[k][i][j]/(0.5*dy[k][i+1][j]+0.5*dy[k][i-1][j]+dy[k][i][j]);
				c24=0.5*cv_zy[k][i][j]/(0.5*dy[k+1][i+1][j]+0.5*dy[k+1][i-1][j]+dy[k+1][i][j]);
				c25=2*cv_zz[k][i][j]/(dz[k+1][i][j]+dz[k][i][j]);


				qxx[k][i][j]=-c1*(head[k][i][j+1]-head[k][i][j])-c2*(head[k][i+1][j+1]-head[k][i-1][j+1])-c3*(head[k][i+1][j]-head[k][i-1][j])
					                                           -c4*(head[k+1][i][j+1]-head[k-1][i][j+1])-c5*(head[k+1][i][j]-head[k-1][i][j]);
				qyy[k][i][j]=-c11*(head[k][i][j+1]-head[k][i][j-1])-c12*(head[k][i+1][j+1]-head[k][i+1][j-1])-c13*(head[k][i+1][j]-head[k][i][j])
					                                           -c14*(head[k+1][i][j]-head[k-1][i][j])-c15*(head[k+1][i+1][j]-head[k-1][i+1][j]);
				qzz[k][i][j]=-c21*(head[k][i][j+1]-head[k][i][j-1])-c22*(head[k+1][i][j+1]-head[k+1][i][j-1])-c23*(head[k][i+1][j]-head[k][i-1][j])
					                                           -c24*(head[k+1][i+1][j]-head[k+1][i-1][j])-c25*(head[k+1][i][j]-head[k][i][j]);
				
			}
}
//---------------------------------------------------------------------------//
//                       output flow ..........                              //
//---------------------------------------------------------------------------//
void Print_flow(int nlay,int nrow,int ncol,string qx_file,float average_qx,vector< vector< vector<float> > > &qxx)
{
	int i,j,k;
	ofstream outFlow;
	outFlow.open(qx_file.c_str());
	
	outFlow<<"flux"<<endl;
	outFlow<<"1"<<endl;
	outFlow<<"Value"<<endl;
	for(k=nlay;k>=1;k--)
	{
	
		for(i=nrow;i>=1;i--)
		{
			for (j=1;j<=ncol;j++)
			{
				outFlow<<qxx[k][i][j]<<endl;
			//	average_qx+=qxx[k][i][j];
			}
		
		}
	
	}
	
	
	outFlow.close();
}
//---------------------------------------------------------------------------//
//                       output head..........                              //
//---------------------------------------------------------------------------//
void Print_head(int nlay,int nrow,int ncol,string head_file,vector< vector< vector<float> > > &head,vector< vector< vector<int> > > ibo)
{
	int i,j,k;
	ofstream outHead;
	outHead.open(head_file.c_str());
	outHead<<"Simulated head"<<endl;
	outHead<<"1"<<endl;
	outHead<<"value"<<endl;
	for(k=nlay;k>=1;k--)
	{
	
		for(i=nrow;i>=1;i--)
		{
			for (j=1;j<=ncol;j++)
			{
				if(ibo[k][i][j]==0)
				{
					head[k][i][j]=-9999;  //  the head of inactive cells are defined as -9999.
				}
				
				outHead<<head[k][i][j]<<endl;
				
			}
			
		}
		
	}	
	outHead.close();
}

//---------------------------------------------------------------------------//
//                      check flows  ..........                              //
//---------------------------------------------------------------------------//
void Check_flow(int nlay,int nrow,int ncol,vector< vector< vector<float> > > &qxx,vector< vector< vector<float> > > &qyy,
		   vector< vector< vector<float> > > &qzz,vector< vector< vector<float> > > &RHS,
		   vector< vector< vector<float> > > &head,vector< vector< vector<float> > >dx,vector< vector< vector<float> > >dy,
			   vector< vector< vector<float> > >dz)
{
	int i,j,k;
	float qtot=0;
	ofstream debug_file;
	debug_file.open("debug.dbg",ios::out|ios::app);
	debug_file<<"row"<<"  "<<"col"<<"  "<<"head[k][i][j]   qxx[k][i][j]   qxx[k][i][j-1]    qyy[k][i][j]    qyy[k][i-1][j]    qzz[k][i][j]   qzz[k-1][i][j]  well_q[k][i][j]  qtot"<<endl;
	for (k=1;k<=nlay;k++)
	{
		debug_file<<"layer"<<k<<endl;
		for (i=1;i<=nrow;i++)
		{
			for (j=1;j<=ncol;j++)
			{
				qtot=qxx[k][i][j]*dy[k][i][j]*dz[k][i][j]-qxx[k][i][j-1]*dy[k][i][j-1]*dz[k][i][j-1]+qyy[k][i][j]*dx[k][i][j]*dz[k][i][j]-qyy[k][i-1][j]*dx[k][i-1][j]*dz[k][i-1][j]
					 +qzz[k][i][j]*dy[k][i][j]*dx[k][i][j]-qzz[k-1][i][j]*dy[k-1][i][j]*dx[k-1][i][j];
			
				debug_file.width(4);
				debug_file<< i;
				debug_file.width(4);
				debug_file<< j;
				debug_file.width(16);
				debug_file.precision(8);
				debug_file.setf(ios_base::right,ios_base::adjustfield);
				debug_file.setf(ios_base::showpoint);
//				ios_base::fmtflags old=debug_file.setf(ios_base::scientific,ios_base::floatfield);
				debug_file<<"  "<<head[k][i][j]<<"   "<<qxx[k][i][j]<<"   "<<qxx[k][i-1][j]<<"   "<<qyy[k][i][j]<<"   "
					<<qyy[k][i][j-1]<<"   "<<qzz[k][i][j]<<"   "<<qzz[k-1][i][j]<<"   "<<RHS[k][i][j]<<"    "<<qtot<<endl;
			}
			debug_file<<endl;
		}
		debug_file<<endl;
	}
	debug_file.close();
}
//-------------------------------------------------------------------------//
//                modify the conductance for constant flow well            //
//-------------------------------------------------------------------------//
void modifyK_well(int nlay,int nrow,int ncol,string &kxx_file,string &kyy_file,string &kzz_file,
				  vector< vector< vector<float> > > &cc_xx,vector< vector< vector<float> > > &cc_xy,
				  vector< vector< vector<float> > > &cc_xz,vector< vector< vector<float> > > &cr_yx,
				  vector< vector< vector<float> > > &cr_yy,vector< vector< vector<float> > > &cr_yz,
				  vector< vector< vector<float> > > &cv_zx,vector< vector< vector<float> > > &cv_zy,
				  vector< vector< vector<float> > > &cv_zz,vector<float>&well_x,vector<float>&well_y,
				  vector<float>&well_z,int n_well,int debug)
{
	
	int i,j,k;
	for (int iw=1;iw<=n_well;iw++){
		for (k=1;k<=nlay;k++)
			for (i=1;i<=nrow;i++)
				for (j=1; j<=ncol;j++)
					if (well_x[iw]==i && well_y[iw]==j &&well_z[iw]==k)
					{
						for (int kk=1;kk<=k;kk++){
							cv_zx[kk][i][j]=99999; 
							cv_zy[kk][i][j]=99999; 
							cv_zz[kk][i][j]=99999; 				
						}
					}
	}
	

}
//-------------------------------------------------------------------------//
//                modify RHS for constant flow rate well                   //
//-------------------------------------------------------------------------//
void modify_RHS_well(int nlay,int nrow,int ncol,vector<float>&well_x,vector<float>&well_y,
				  vector<float>&well_z,vector<float>&well_q,int n_well,vector< vector< vector<float> > > &RHS)
{int i,j,k;
	for (int iw=1;iw<=n_well;iw++)
		for (k=1;k<=nlay;k++)
			for (i=1;i<=nrow;i++)
				for (j=1; j<=ncol-1;j++)
				{
					if (well_x[iw]==i && well_y[iw]==j &&well_z[iw]==k)
						RHS[k][i][j]=RHS[k][i][j] - well_q[iw];
				}
}